from __future__ import annotations

# Std lib
from abc import ABC, abstractmethod
from typing import List, Unpack, Tuple, Literal, ClassVar, Type
from warnings import warn

# 3rd party lib
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.animation import FuncAnimation
from matplotlib.patches import PathPatch, Circle
from matplotlib.path import Path
from numpy.typing import NDArray
from shapely import Polygon, MultiPolygon, GeometryCollection, unary_union

# Local lib
from .. import shapes
from ..beams import BeamType
from ..mill import Mill
from ..resources import TIMEUNITS, length_units, Material, Materials, LENGTHUNITS, time_units
from ..scan_styles import ScanStyles, ScanStyle

BOUNDARIES = Literal["x_min", "x_max", "y_min", "y_max"]


class FIBInterface(ABC):
    _fov_pixels: ClassVar[int]

    @abstractmethod
    def get_pixelsize(self, units: str) -> float:
        """Returns the pixelsize in the specified units for the current FIB settings."""
        ...

    @abstractmethod
    def get_fov(self, units: str) -> float:
        ...

    @property
    @abstractmethod
    def beam(self) -> BeamType:
        ...


class Add:
    # region Class Body

    _pattern: Pattern

    # endregion Class Body

    # region Dev Methods

    def __init__(self, pattern: Pattern):
        self._pattern = pattern

    # endregion Dev Methods

    # region User Methods

    def rectangle(self, **kwargs: Unpack[shapes.RectangleKwargs]) -> shapes.Rectangle:
        rect = shapes.Rectangle(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(rect)
        return rect

    def circle(self, **kwargs: Unpack[shapes.CircleKwargs]) -> shapes.Circle:
        circ = shapes.Circle(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(circ)
        return circ

    def polygon(self, **kwargs: Unpack[shapes.PolyKwargs]) -> shapes.Poly:
        poly = shapes.Poly(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(poly)
        return poly

    def regular_polygon(self, **kwargs: Unpack[shapes.RegularPolygonKwargs]) -> shapes.RegularPolygon:
        poly = shapes.RegularPolygon(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(poly)
        return poly

    def lattice(self, **kwargs: Unpack[shapes.LatticeKwargs]) -> shapes.Lattice:
        lattice = shapes.Lattice(pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(lattice)
        return lattice

    def text(self, txt: str, **kwargs: Unpack[shapes.TextKwargs]) -> shapes.Text:
        text = shapes.Text(txt, pattern=self._pattern, **kwargs)  # type: ignore
        self._pattern._add_shape(text)
        return text
    # endregion User Methods


class Pattern:

    # region Class Body

    _shapes: List[shapes.Shape]
    _mill: Mill
    _fib: FIBInterface
    _x: float
    _y: float
    _pass_multiplier: int
    _scan_style: ScanStyle
    _materials: List[Material]
    Materials: Type[Materials]
    ScanStyles: Type[ScanStyles]
    add: Add

    __slots__ = ["_shapes", "_mill", "_fib", "_x", "_y", "_pass_multiplier", "add", "_scan_style", "_materials",
                 "Materials", "ScanStyles"]

    # endregion Class Body

    # region Dev Methods

    def __init__(self, fib, x: float = 0, y: float = 0) -> None:
        self._fib = fib
        self._shapes = []
        self._mill = Mill()
        self.add = Add(self)
        self.set_position(x, y)
        self.ScanStyles = ScanStyles
        self._scan_style = ScanStyles.Contour.inwards
        self._materials = [Materials.Default_Si()]
        self.Materials = Materials

    def _add_shape(self, shape: shapes.Shape) -> None:
        self._shapes.append(shape)

    def _get_millable_area(self) -> Polygon | MultiPolygon | GeometryCollection:

        millable = []
        non_millable = []
        override_millable = []

        def filter_millable(shape_to_filter: shapes.Shape) -> None:
            """Recursively filter all shapes based on the milling flag."""
            if isinstance(shape_to_filter, shapes.CombinedShape):
                for shape in shape_to_filter._shapes:
                    filter_millable(shape)
            elif isinstance(shape_to_filter, shapes.Lattice):
                for row in shape_to_filter._shapes:
                    for shape in row:
                        if shape is not None:
                            filter_millable(shape)
            elif isinstance(shape_to_filter, shapes.Shape):
                if shape_to_filter._millable:
                    if shape_to_filter._millable_override:
                        override_millable.append(shape_to_filter.polygon)
                    else:
                        millable.append(shape_to_filter.polygon)
                else:
                    non_millable.append(shape_to_filter.polygon)
            else:
                raise TypeError(f"Got '{type(shape_to_filter)}', expected 'Shape' or any of it's subclasses.")
            return

        for pattern_shape in self._shapes:
            filter_millable(pattern_shape)

        millable_area = unary_union(millable)
        non_millable_area = unary_union(non_millable)
        millable_area = millable_area.difference(non_millable_area)
        millable_area = unary_union([millable_area, *override_millable])

        return millable_area

    def _get_pattern_ax(self, ax: plt.Axes = None, alpha: float = 1) -> plt.Axes:
        """Makes a plot where the millable area is filled in, not including any area containing a non-millable shape."""

        millable_area = self._get_millable_area()

        if ax is None:
            ax = plt.gca()

        if millable_area.is_empty:
            print("No millable area to plot.")
        else:
            if millable_area.geom_type == "MultiPolygon":
                for poly in millable_area.geoms:
                    exterior_coords = np.array(poly.exterior.coords)
                    interiors = [np.array(interior.coords) for interior in poly.interiors]
                    path = Path.make_compound_path(Path(exterior_coords), *[Path(interior) for interior in interiors])
                    patch = PathPatch(path, facecolor='black', edgecolor='none', alpha=alpha, label="Millable")
                    ax.add_patch(patch)
            else:
                poly = millable_area
                exterior_coords = np.array(poly.exterior.coords)
                interiors = [np.array(interior.coords) for interior in poly.interiors]
                path = Path.make_compound_path(Path(exterior_coords), *[Path(interior) for interior in interiors])
                patch = PathPatch(path, facecolor='black', edgecolor='none', alpha=alpha, label="Millable")
                ax.add_patch(patch)

        return ax

    def _pixelate_beam_path(self, streamlines: NDArray) -> NDArray:

        # Fetch the pixel fov
        pixel_fov = self._fib._fov_pixels

        # Conversion factor: nm -> pixels
        conversion_factor = 1 / self._fib.get_pixelsize("nm")

        # Extract physical positions (assuming column 1 = x, column 2 = y)
        streamlines[:, 1] = np.rint(streamlines[:, 1] * conversion_factor)
        streamlines[:, 2] = np.rint(streamlines[:, 2] * conversion_factor)

        # Readjust overal position to position in pixel space
        streamlines[:, 2] *= -1
        streamlines[:, 1:3] += np.array([pixel_fov // 2, pixel_fov // 2])

        # Check if the pattern is within the field of view
        x = (streamlines[:, 1] < 0) | (streamlines[:, 1] >= self._fib._fov_pixels)
        y = (streamlines[:, 2] < 0) | (streamlines[:, 2] >= self._fib._fov_pixels)

        # Check if the y-dimension is within the recomended pixel range:
        y_recomended = (streamlines[:, 2] < 4479) | (streamlines[:, 2] > 61055)  # Lowest/highest recomemded y-pixels.

        if x.any() or y.any():  # Use .any() to check if any elements are True
            raise ValueError(f"There are millable shapes outside of the field of view, which is currently "
                             f"{self._fib.get_fov('um')}x{self._fib.get_fov('um')} microns.")
        elif y_recomended.any():
            warn(f"\nWarning!: \nThere are millable areas falling outside the recomended pixel coordinates in the "
                 f"y-axis, which is 4479 - 61055.\nMilling might still work, but it's possible that areas outside this "
                 f"range might be treated unexptectedly. Caution is adviced.")

        return streamlines

    def _get_passes(self, streamlines: NDArray = None) -> int:

        BEAM_CURRENT = self._fib.beam.get_current("na")
        AREA = length_units(self._get_millable_area().area, "nm", "um", power=2)
        total_milling_depth = self._mill.get_mill_depth("um")

        time_goal = 0
        for material in self._materials:

            # Fetch the thickness of the current material.
            thickness = material.get_thickness("um")

            # If material is thicker than milling depth, use milling depth, if not, use material thickness.
            this_milling_depth = min(thickness, total_milling_depth)

            # Find volume for milling this material
            millable_volume = AREA * this_milling_depth

            # Subtract the milled depth from the total milling depth
            total_milling_depth -= this_milling_depth

            # Add the time for milling this material to the total time
            time_goal += millable_volume / (BEAM_CURRENT * material.volume_pr_dose)

        effective_mill_time = 0
        for vec in streamlines:
            if vec.shape[0] == 4:
                if vec[3] == 0:
                    continue
            effective_mill_time += vec[0]
        effective_mill_time = time_units(effective_mill_time * 100, "ns", "s")

        nr_passes = np.rint(time_goal/effective_mill_time).astype(int)

        return nr_passes

    # endregion Dev Methods

    # region User Methods

    def set_scan_style(self, scan_style: ScanStyle):
        """Sets the scan style to be used when generating the beam path. Scan styles are found in self.ScanStyles."""
        if not isinstance(scan_style, ScanStyle):
            raise TypeError(f"Expected type ScanStyle, got '{type(scan_style)}'.")
        self._scan_style = scan_style

    def set_material(self, *material: Material | Type[Material]) -> None:
        """
        Sets the materials the sample is made out of.

        If more than one material is specified, the first material will be the top layer, and subsequent materials are
        layered beneath. Materials can be found under the Materials variable of the pattern instance.

        The thickness of each material is specified by calling the material with thickness provided like this:

        pattern.Materials.Au(100, "nm")

        If a thickness is not specified, the thickness is by default infinite, for example:

        pattern.Materials.Au

        The bottom layer will allways have inifinite thickmness no matter what thickness is specified.


        Example usage:

        # Sample consisting of an infinetely thick gold layer.
        pattern.set_material(pattern.Materials.Au)

        # Sample consisting of a 100 nm gold layer with an infinitely thick layer of Silicon beneath.
        pattern.set_material(pattern.Materials.Au(100, "nm"), pattern.Materials.Si)

        # If more than one material is specified, and a layer that is not the bottom layer has no specified thickness,
        you will get a ValueError, ie. like this:
        pattern.set_material(pattern.Materials.Au, pattern.Materials.Si)
        """
        materials: List[Material] = []

        for mat in material:
            if isinstance(mat, Material):
                materials.append(mat)
            elif isinstance(mat, type) and issubclass(mat, Material):
                materials.append(mat())  # Instantiate if it's a class
            else:
                raise TypeError(f"Expected type 'Material' or subclass, got '{type(mat)}'.")

        setattr(materials[-1], "_thickness", float('inf'))

        if len(materials) > 1:
            for mat in materials[:-1]:
                if mat.get_thickness() == float('inf'):
                    raise ValueError(f"A material that is not the bottom layer has infinite thickness.")

        self._materials = materials

    def write_stream(self, filename: str) -> None:
        streamlines = np.vstack(self._scan_style.value(self._get_millable_area(), self._mill, self._fib.beam, True))
        streamlines = self._pixelate_beam_path(streamlines)
        passes = self._get_passes(streamlines)

        if not filename.endswith(".str"):
            filename += ".str"
        with open(filename, "w", encoding="utf-8") as file:
            file.writelines(["s16\n", str(passes) + "\n",  str(streamlines.shape[0]) + "\n"])
            file.writelines([
                f"{int(line[0])} {int(line[1])} {int(line[2])}"
                f"{' ' + str(int(line[3])) if line[3] != 2 else ''}\n"
                for line in streamlines
            ])
        return

    def set_position(self, x: float = 0, y: float = 0, units: LENGTHUNITS = "nm") -> None:
        """Sets the position of the pattern, including all contained shapes."""
        self._x, self._y = length_units(x, units, "nm"), length_units(y, units, "nm")

    def plot_pattern(self, figsize: Tuple[float, float] = (15, 10), alpha: float = 0.6) -> None:
        """Makes a plot where the millable area is filled in, not including any area containing a non-millable shape."""
        plt.rcParams["figure.figsize"] = figsize
        prev_pos = self._x, self._y
        self.set_position(0, 0)
        ax = self._get_pattern_ax(None, alpha)
        self.set_position(*prev_pos)
        ax.relim()
        ax.autoscale_view()
        ax.set_aspect('equal')
        # Retrieve the current handles and labels
        handles, labels = plt.gca().get_legend_handles_labels()

        # Create a dictionary to remove duplicates. In Python 3.7+, dict preserves insertion order.
        by_label = dict(zip(labels, handles))

        # Now pass the unique handles and labels to plt.legend()
        plt.legend(by_label.values(), by_label.keys())
        plt.show()

    def plot_beam_path(self, figsize: Tuple[float, float] = (15, 10), alpha: float = 0.25,
                       rasterize: bool = True) -> None:

        millable_area = self._get_millable_area()
        lines = self._scan_style.value(millable_area, self._mill, self._fib.beam, rasterize)

        plt.figure(figsize=figsize)
        ax = self._get_pattern_ax(None, alpha)
        for line in lines:
            plt.plot(line[:, 1], line[:, 2], "o-")

        plt.gca().set_aspect("equal")
        plt.show()

    def animate_beam_path(self, figsize=(15, 10), alpha=0.25, repeat: bool = False, step: int = 30,
                          interval: float = 5) -> None:
        """
        Animates drawing the beam path contours with fixed axis limits.
        This version updates pre-created artists using blitting to improve performance.
        """
        # Get the overall geometry and compute the contour lines.
        millable_area = self._get_millable_area()
        lines = self._scan_style.value(millable_area, self._mill, self._fib.beam, True)

        # Precompute coordinate data for each line.
        # Each entry in line_data is a tuple (x_values, y_values)
        line_data = []
        for line in lines:
            x, y = line[:, 1], line[:, 2]
            line_data.append((list(x), list(y)))

        # Compute the overall bounding box from all lines.
        all_x = [x for (x_vals, _) in line_data for x in x_vals]
        all_y = [y for (_, y_vals) in line_data for y in y_vals]
        margin = 1  # extra space around the data
        x_min, x_max = min(all_x) - margin, max(all_x) + margin
        y_min, y_max = min(all_y) - margin, max(all_y) + margin

        # Create the figure and axis with fixed limits.
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_xlim(x_min - 50, x_max + 50)
        ax.set_ylim(y_min - 50, y_max + 50)
        ax.set_aspect("equal")
        self._get_pattern_ax(ax, alpha=alpha)

        # Pre-create a line artist for the current (active) beam path.
        current_line_artist, = ax.plot([], [], color='red')

        # Pre-create a circle to indicate the beam's current position.
        beam_radius = self._fib.beam.get_radius("nm")
        beam_circle = Circle((0, 0), beam_radius, color='black', fill=True, alpha=0.5, linewidth=0)
        ax.add_patch(beam_circle)

        # Pre-create a container for completed line artists.
        completed_line_artists = []

        # Animation state dictionary.
        state = {
            'current_line': 0,  # index of the line being drawn
            'current_point': 0  # index of the next point in the current line
        }

        # Determine the total number of frames (one per point).
        total_frames = sum(len(x_vals) for (x_vals, _) in line_data)

        def init():
            """Initialize the artists."""
            current_line_artist.set_data([], [])
            beam_circle.center = (0, 0)
            return [current_line_artist, beam_circle]

        def update(frame):
            """Update function for FuncAnimation."""
            if state['current_line'] < len(line_data):
                x_vals, y_vals = line_data[state['current_line']]
                # Handle single point lines separately.
                if len(x_vals) == 1:
                    # Immediately complete the single-point line using a marker.
                    current_line_artist.set_data(x_vals, y_vals)
                    beam_circle.center = (x_vals[0], y_vals[0])
                    completed_line, = ax.plot(x_vals, y_vals, 'o', markersize=1, color='blue', alpha=0.3)
                    completed_line_artists.append(completed_line)
                    state['current_line'] += 1
                    state['current_point'] = 0
                    current_line_artist.set_data([], [])
                else:
                    # Increase the number of points drawn for the current line.
                    state['current_point'] += step
                    if state['current_point'] >= len(x_vals):
                        # Draw the full current line in a static (completed) style.
                        current_line_artist.set_data(x_vals, y_vals)
                        beam_circle.center = (x_vals[-1], y_vals[-1])
                        completed_line, = ax.plot(x_vals, y_vals, color='blue', alpha=0.3)
                        completed_line_artists.append(completed_line)
                        state['current_line'] += 1
                        state['current_point'] = 0
                        current_line_artist.set_data([], [])
                    else:
                        # Update the current line artist with the new data.
                        current_line_artist.set_data(x_vals[:state['current_point']],
                                                     y_vals[:state['current_point']])
                        beam_circle.center = (x_vals[state['current_point'] - 1],
                                              y_vals[state['current_point'] - 1])
            return [current_line_artist, beam_circle] + completed_line_artists

        # Create the animation with blitting enabled.
        anim = FuncAnimation(
            fig, update, init_func=init, frames=total_frames, interval=interval,
            blit=True, repeat=repeat, cache_frame_data=True
        )

        plt.show()

    def get_dwell_time(self, units: TIMEUNITS) -> float:
        return self._mill.get_dwell_time(units)

    def set_dwell_time(self, dt: float, units: TIMEUNITS) -> None:
        self._mill.set_dwell_time(dt, units)

    def set_depth(self, depth: float | List[Material | float] | Material, units: LENGTHUNITS = "um") -> None:
        d = 0

        if isinstance(depth, list):
            if not any([isinstance(material, (Material, float, int)) for material in depth]):
                raise TypeError(f"When a list is passed to this method, elements should be of type 'Material' "
                                f"or a subclass, or a number (int or float).")
            for material in depth:
                if isinstance(material, Material):
                    mat_d = material.get_thickness(units)
                    if mat_d == float('inf'):
                        raise ValueError(
                            f"Material '{material}' has inifite thickness, which yields an infinite milling depth. "
                            f"For obvious reasons, you can't set the milling depth using this material.")
                    d += mat_d

                else:
                    d += material

        elif isinstance(depth, Material):
            mat_d = depth.get_thickness("um")
            if mat_d == float('inf'):
                raise ValueError(f"Material '{depth}' has inifite thickness, which yields an infinite milling depth. "
                                 f"For obvious reasons, you can't set the milling depth using this material.")
            d += mat_d

        elif isinstance(depth, (int, float)):
            d += depth
        else:
            raise TypeError(f"Expected type 'Material' or a subclass, a number (int or float), or a list of "
                            f"Material/numbers. Got '{type(depth)} = {depth}'.")

        self._mill.set_mill_depth(d, units)

    def get_depth(self, units: LENGTHUNITS) -> float:
        return self._mill.get_mill_depth(units)

    def get_overlap(self) -> float:
        return self._mill.get_overlap()

    def set_overlap(self, overlap_fraction: float) -> None:
        self._mill.set_overlap(overlap_fraction)

    def set_pass_multiplier(self, multiplier: int) -> None:
        """ Sets the number of hardcoded passes for the beam path.

            - If set to 1, the beam will trace the pattern once in the exported stream file.
              The number of actual passes performed will then be determined by the "nr. of passes" parameter in the
              stream file or an override in the FIB machine's interface.
            - If set to ie. 2, the beam will trace the pattern twice in the exported stream file.
              This is equivalent to setting the multiplier to 1 and adjusting the "nr. of passes" parameter to 2.

            This feature is particularly useful for parameter sweeps. For example, suppose you need to create two
            versions of the same pattern with different dwell times while using the same stream file.
            To achieve the same milling depth, the number of passes must vary. However, since the stream file
            enforces a single pass setting for the entire file, you can work around this by adjusting
            the pass multiplier.

            For instance, if one pattern has a 1 µs dwell time and another has 0.5 µs, setting the pass multiplier of
            the 0.5 µs pattern to 2 ensures both receive the same dose.
        """
        # assert isinstance(multiplier, int)
        # self._pass_multiplier = multiplier
        raise NotImplementedError(f"This feature is currently not supported. Will be in later versions.")

    def place_next_to(self, pattern: Pattern, boundary: BOUNDARIES, offset: float = 0) -> None:
        """Places the pattern adjacent to the specified boundary of another pattern."""
        direction = -1 if "min" in boundary else 1
        pos = getattr(pattern, boundary) + (getattr(self, f"{boundary[0]}_span") / 2) * direction + offset
        self.set_position(**{boundary[0]: pos})

    # endregion User Methods

    # region Properties

    @property
    def substrate(self) -> List[Material]:
        """Returns a list of the materials of the pattern substrate, top to bottom."""
        return self._materials

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, x: float) -> None:
        self.set_position(x=x)

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, y: float) -> None:
        self.set_position(y=y)

    @property
    def x_span(self) -> float:
        return self.x_max - self.x_min

    @property
    def y_span(self) -> float:
        return self.y_max - self.y_min

    @property
    def x_min(self) -> float:
        return min([shape.x_min for shape in self._shapes]) + self._x

    @property
    def x_max(self) -> float:
        return max([shape.x_max for shape in self._shapes]) + self._x

    @property
    def y_min(self) -> float:
        return min([shape.y_min for shape in self._shapes]) + self._y

    @property
    def y_max(self) -> float:
        return max([shape.y_max for shape in self._shapes]) + self._y

    # endregion Properties
